theory Aexp
imports Main
begin


(* - 'var_type : type des variables du programme.
   
  Les variables symboliques sont des versions indicées des variables *)
type_synonym 'var_type symvar = "'var_type \<times> nat"

type_synonym D = integer

(* Un etat d'un programme est une fonction (totale) associant des entiers 
   aux variables du programme *)
type_synonym 'var_type state = "'var_type \<Rightarrow> D"

(* Une expression arithmetique est une fonction associant des entiers aux
   etats *)
type_synonym 'var_type aexp =  "'var_type state \<Rightarrow> D"

(* ensemble des variables d'une expression = ensemble des variables ayant une influence sur la 
   valeur associée par l'expression à un état quelconque *)
definition vars :: "'var_type aexp \<Rightarrow> 'var_type set" where
          "vars exp = {v. \<exists> \<sigma> val. exp (\<sigma>(v := val)) \<noteq> exp \<sigma>}"

(* une variable est fraiche pour une expression si elle ne figure pas dans l'ensemble des variables 
   de cette expression *)
abbreviation fresh :: "'var_type \<Rightarrow> 'var_type aexp \<Rightarrow> bool" where
            "fresh v exp \<equiv> v \<notin> vars exp"



(***********************************************************)
(******************* Théroèmes *****************************)
(***********************************************************)

(* l'ensemble des variables symboliques est infini *)
lemma infinite_symvars :
  "\<not> finite (UNIV::'var_type symvar set)"
by (simp add : finite_prod)

(* consequence de la "fraicheur" d'une variable pour une expression *) 
lemma fresh_is :
  "fresh v exp \<Longrightarrow> exp (\<sigma>(v := val)) = exp \<sigma>"
by (simp add : vars_def)


find_unused_assms

end