theory SubExt
imports SubRel
begin


(**************************************************************************************************)
(***************************  EXTENSION PAR AJOUT D'UNE SUBSOMPTION  ******************************)
(**************************************************************************************************)

(* On définit d'abord l'extension par ajout d'une subsomption. Etant donnés un graphe g, une 
   subsomption (d,a) (dans la définition, (d,a) est représenté par sub) et une relation de 
   subsomption subs, subs' étend subs pour g si :
   - d \<noteq> a,
   - d et a sont des versions indicées d'un même sommet,
   - d et a sont des sommets de g (qui est un graphe dont les sommets sont indicés, ie. la partie
     rouge d'un RB-graphe),
   - d n'est pas déjà subsumé dans subs et ne subsume pas non plus,
   - a n'est pas lui même déjà subsumé,
   - (dans la partie rouge,) aucun arc n'est issu de d, et
   - subs' est simplement subs à laquelle on a ajouté (d,a).

*)


definition extends ::
  "(('a \<times> nat),'b) rgraph_scheme \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat))     \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> bool"
where
  "extends g subs sub subs' \<equiv> (
    (*sub_rel_of g subs                                    \<and>*)
    subsumee sub \<noteq> subsumer sub                          \<and>
    fst (subsumee sub)  = fst (subsumer sub) \<and>
    subsumee sub \<in> Graph.vertices g                           \<and>
    subsumee sub \<notin> subsumers subs                        \<and>
    subsumee sub \<notin> subsumees subs                        \<and>
    subsumer sub \<in> Graph.vertices g                           \<and>
    subsumer sub \<notin> subsumees subs                        \<and>
    out_arcs g (subsumee sub) = {}                       \<and>
    subs' = subs \<union> {sub}
  )"

(*definition extends2 :: 
  "(('a \<times> nat),'b) rgraph_scheme \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> bool"
where
  "extends2 g subs subs' \<equiv> \<exists> sub. subs' = insert sub subs
                                \<and> out_arcs g (subsumee sub) = {}
                                \<and> subsumee sub \<noteq> subsumer sub
                                \<and> fst (subsumee sub)  = fst (subsumer sub)
                                \<and> subsumee sub \<in> Graph.vertices g
                                \<and> subsumee sub \<notin> subsumers subs
                                \<and> subsumee sub \<notin> subsumees subs
                                \<and> subsumer sub \<in> Graph.vertices g 
                                \<and> subsumer sub \<notin> subsumees subs"*)

(*inductive extends2 ::
  "(('a \<times> nat),'b) rgraph_scheme \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat))     \<Rightarrow> bool"
where
  "subsumee sub \<noteq> subsumer sub \<Longrightarrow>
   fst (subsumee sub) = fst (subsumer sub) \<Longrightarrow>
   out_arcs g (subsumee sub) = {} \<Longrightarrow>
   subsumee sub \<in> Graph.vertices g \<Longrightarrow>
   subsumee sub \<notin> subsumers subs \<Longrightarrow>
   subsumee sub \<notin> subsumees subs \<Longrightarrow>
   subsumer sub \<in> Graph.vertices g \<Longrightarrow>
   subsumer sub \<notin> subsumees subs \<Longrightarrow>
   extends2 g subs sub"*)

lemma extendsE1 :
  assumes "extends g subs sub subs'"
  shows   "sub \<in> subs'"
using assms by (simp add : extends_def)

lemma extendsE2 :
  assumes "extends g subs sub subs'"
  shows   "subs' = insert sub subs"
using assms by (simp add : extends_def)

lemma extendsE3 :
  assumes "extends g subs sub subs'"
  shows   "subsumee sub \<noteq> subsumer sub"
using assms by (simp add : extends_def)

lemma extends_imp_not_mem :
  assumes "extends g subs sub subs'"
  
  shows   "sub \<notin> subs"
proof (rule ccontr, subst (asm) not_not)

  assume "sub \<in> subs"

  hence  "(fst sub,snd sub) \<in> subs" by simp

  hence  "subsumee sub \<in> subsumees subs" unfolding subsumees_conv by blast

  thus "False" using assms by (simp add : extends_def)
qed


(* L'extension d'une relation de subsomption appartenant à un graphe g appartient également à g *)
lemma (in sub_rel_of) extends_imp_sub_rel_of :
  assumes "extends g subs sub subs'"
  
  shows   "sub_rel_of g subs'"
using assms sub_rel_of 
unfolding extends_def sub_rel_of_def subsumers_conv subsumees_conv 
by (auto simp del : split_paired_Ex)

(*lemma (in sub_rel_of) extends_imp_sub_rel_of :
  assumes "extends2 g subs subs'"
  
  shows   "sub_rel_of g subs'"
using assms sub_rel_of
by (auto simp add : subsumees_conv subsumers_conv sub_rel_of_def extends2_def)*)


(* L'extension d'une relation de subsomption bien formée est bien formée *)
lemma (in wf_sub_rel) extends_imp_wf_sub_rel :
  assumes "extends g subs sub subs'"
  shows   "wf_sub_rel subs'"
unfolding wf_sub_rel_def
proof (intro conjI)

  show "subsumers subs' \<inter> subsumees subs' = {}"
  using wf_sub_rel assms by (auto simp add : extends_def subsumers_conv subsumees_conv)

next

  show "\<forall>sub. sub \<in> subs' \<longrightarrow> fst (subsumee sub) = fst (subsumer sub)"
  using wf_sub_rel assms unfolding extends_def by auto

next

  show "\<forall>d\<in>subsumees subs'. \<exists>!a. (d, a) \<in> subs'"
  unfolding Ball_def
  proof (intro allI impI)

    fix d

    assume "d \<in> subsumees subs'" 

    hence  "d = subsumee sub \<or> d \<in> subsumees subs"
    using extendsE2[OF assms] unfolding subsumees_conv by auto

    thus "\<exists>! a. (d,a) \<in> subs'"
    proof (elim disjE)
      
      case goal1

      show ?thesis
      unfolding Ex1_def
      proof (rule_tac ?x="subsumer sub" in exI, intro conjI)

        show "(d, subsumer sub) \<in> subs'" 
        using goal1 extendsE2[OF assms] by simp

      next

        have "d \<notin> subsumees subs" 
        using assms goal1 by (auto simp add : extends_def)
        
        thus "\<forall> a. (d, a) \<in> subs' \<longrightarrow> a = subsumer sub" 
        using extendsE2[OF assms] unfolding subsumees_conv
        by (auto simp del : split_paired_All split_paired_Ex)
        
      qed
      
    next

      case goal2

      then obtain a where "(d,a) \<in> subs" unfolding subsumees_conv by blast

      hence "d \<noteq> subsumee sub" 
      using assms unfolding extends_def subsumers_conv subsumees_conv
      by (force simp del : split_paired_All split_paired_Ex)

      show ?thesis 
      unfolding Ex1_def
      proof (rule_tac ?x="a" in exI, intro conjI)
      
        show "(d, a) \<in> subs'" using `(d,a) \<in> subs` extendsE2[OF assms] by simp

      next

        show "\<forall> a'. (d, a') \<in> subs' \<longrightarrow> a' = a"
        proof (intro allI impI)

          fix a'

          assume "(d,a') \<in> subs'"

          hence  "(d,a') \<in> subs" using `d \<noteq> subsumee sub` extendsE2[OF assms] by auto

          thus "a' = a"
          using `(d,a) \<in> subs` subsumed_by_one unfolding subsumees_conv Ex1_def 
          by (auto simp del : split_paired_All split_paired_Ex)

        qed
      qed
    qed
  qed
qed


lemma (in wf_sub_rel_of) extends_imp_wf_sub_rel_of :
  assumes "extends g subs sub subs'"
  shows   "wf_sub_rel_of g subs'"
using sub_rel_of.extends_imp_sub_rel_of[OF sub_rel_of assms] 
      wf_sub_rel.extends_imp_wf_sub_rel[OF wf_sub_rel assms]
by (simp add : wf_sub_rel_of_def)

(*unfolding wf_sub_rel_of_def
proof (intro conjI)

  show "sub_rel_of g subs'" 
  using assms sub_rel_of 
  unfolding extends_def sub_rel_of_def subsumers_conv subsumees_conv
  by (auto simp del : split_paired_Ex)
 

next

  show "wf_sub_rel subs'"
  unfolding wf_sub_rel_def
  proof (intro conjI)
    
    show "subsumers subs' \<inter> subsumees subs' = {}"
    using assms wf_sub_rel wf_sub_rel.inter_empty[of subs]
    unfolding extends_def subsumers_conv subsumees_conv
    by (auto simp del : split_paired_Ex)

  next
    
    show "\<forall> sub. sub \<in> subs' \<longrightarrow> fst (subsumee sub) = fst (subsumer sub)"
    using assms wf_sub_rel wf_sub_rel.sub_imp_same_verts by (auto simp add : extends_def)

  next

    show "\<forall> d \<in> subsumees subs'. \<exists>!a. (d, a) \<in> subs'"
    unfolding Ball_def
    proof (intro allI impI)

      fix d'

      assume "d' \<in> subsumees subs'" 

      moreover
      have   "sub \<notin> subs" using assms by (rule extends_imp_not_mem)

      ultimately
      have "\<exists> a'. (d',a') = sub \<or> ((d',a') \<noteq> sub \<and> (d',a') \<in> subs)"
      using assms by (auto simp add : extends_def subsumers_conv subsumees_conv)

      then obtain a' where "(d',a') = sub \<or> ((d',a') \<noteq> sub \<and> (d',a') \<in> subs)" by blast

      thus "\<exists>! a. (d',a) \<in> subs'"
      proof (elim disjE)

        assume "(d', a') = sub"

        moreover
        hence  "d' \<notin> subsumees subs" 
        using assms by (auto simp add : extends_def subsumers_conv subsumees_conv)

        ultimately
        show ?thesis using assms unfolding extends_def Ex1_def subsumers_conv subsumees_conv
        by (simp del : split_paired_All split_paired_Ex) blast

      next
      
        assume "(d', a') \<noteq> sub \<and> (d', a') \<in> subs" (is "?lh \<and> ?rh")

        moreover
        hence  "d' \<noteq> subsumee sub" 
        using assms subsumees_conv unfolding extends_def by blast
        
        ultimately
        show ?thesis 
        using assms wf_sub_rel wf_sub_rel.subsumed_by_one[of subs] 
        unfolding extends_def Ex1_def subsumers_conv subsumees_conv
        by (rule_tac ?x="a'" in exI) (auto simp del : split_paired_All split_paired_Ex)
      qed
    qed
  qed
qed*)





(* un sous-chemin d'un graphe muni d'une relation de subsomption est également un sous-chemin
   du même graphe dont la relation de subsomption a été étendue *)
lemma sp_in_extends :
  assumes "extends g subs sub subs'"
  assumes "subpath g v1 p v2 subs"
  
  shows   "subpath g v1 p v2 subs'"
using assms unfolding extends_def subpath_def
by (auto simp only : cas_Un)

lemma sp_from_in_extends :
  assumes "extends g subs sub subs'"
  assumes "subpath_from g v p subs"
  
  shows   "subpath_from g v p subs'"
using assms sp_in_extends unfolding subpath_from_def by blast

lemma p_in_extends :
  assumes "extends g subs sub subs'"
  assumes "path g p v subs"
  shows   "path g p v subs'"
using assms unfolding path_def by (rule sp_in_extends) 



lemma (in wf_sub_rel_of) extends_imp_sp_appendI :
  assumes "extends g subs (d,a) subs'"
  assumes "subpath g v1 p1 d subs"
  assumes "subpath g a p2 v2 subs'"

  shows   "subpath g v1 (p1 @ p2) v2 subs'"
proof -

  have "wf_sub_rel subs'" 
  proof -
    have "wf_sub_rel_of g subs'"
    using assms(1) wf_sub_rel_of by (auto simp add : extends_imp_wf_sub_rel_of)

    thus ?thesis by (simp add : wf_sub_rel_of_def)
  qed

  moreover
  have "(d,a) \<in> subs'" using assms(1) by (simp add : extends_def)

  moreover
  have "subpath g v1 p1 d subs'" using assms(1) assms(2) by (rule sp_in_extends)

  ultimately
  show ?thesis using assms(3) wf_sub_rel by (intro wf_sub_rel.sp_appendI2)
qed


(* Soient v1 et v2 des sommets de g reliés par le chemin vide dans l'extension (par subsomp\<degree>) de g.
   
   v1 et v2 sont également reliés par le chemin vide dans g si et seulement si ces sommets ne 
   correspondent pas aux extrémités de la nouvelle subsomption. *)

lemma (in wf_sub_rel_of) sp_Nil_in_extends :
  assumes "extends g subs (d,a) subs'"
  assumes "subpath g v1 [] v2 subs'"

  shows   "subpath g v1 [] v2 subs \<longleftrightarrow> (v1 \<noteq> d \<or> v2 \<noteq> a)"
          (is "?L \<longleftrightarrow> ?R")
proof (intro iffI)

  assume "?L"
  
  show   "?R"
  proof (rule ccontr, simp only : de_Morgan_disj not_not, elim conjE)
  
    assume "v1 = d" and "v2 = a"
  
    hence  "d = a \<or> (d,a) \<in> subs\<^sup>+" using `?L` by (auto simp add : subpath_def)
  
    thus "False"
    proof (elim disjE)
      assume "d = a" thus ?thesis using assms (1) by (simp add : extends_def)
    next
      assume "(d,a) \<in> subs\<^sup>+"
      hence  "(d,a) \<in> subs" by (rule in_trancl_imp)
      thus ?thesis using assms(1) extends_imp_not_mem[of g subs "(d,a)" subs'] by simp
    qed
  qed
  
next

  assume "?R"

  have 1 : "v1 = v2 \<or> (v1,v2) \<in> subs'\<^sup>+" 
  and  2 : "v1 \<in> Graph.vertices g" 
  using assms(2) by (simp_all add : subpath_def)
  
  have 3 : "v2 \<in> Graph.vertices g"
  proof -
    have "sub_rel_of g subs'" 
    using sub_rel_of assms(1) by (rule sub_rel_of.extends_imp_sub_rel_of)

    thus ?thesis using assms(2) by (rule sub_rel_of.lst_of_sp_is_vert)
  qed
  
  have "v1 = v2 \<or> (v1,v2) \<in> subs\<^sup>+"
  using 1
  proof (elim disjE)
    assume "v1 = v2" thus ?thesis by simp
  next
    assume "(v1, v2) \<in> subs'\<^sup>+"

    moreover
    have   "wf_sub_rel_of g subs'" using assms(1) by (rule extends_imp_wf_sub_rel_of)

    ultimately
    have   "(v1,v2) \<in> subs'" by (elim wf_sub_rel_of.in_trancl_imp)

    thus ?thesis 
    using assms(1) `?R` by (auto simp add : extends_def)
  qed

  thus "?L" using 2 3 by (simp add : subpath_def)
qed


(*************************** passage par une subsomption ******************************************)

(* uses_sub v1 as v2 sub vaut vrai si as nécessite la subsumption sub pour être consistante.
*)

fun uses_sub :: 
  "('a \<times> nat)                \<Rightarrow> 
   ('a \<times> nat) arc list       \<Rightarrow> 
   ('a \<times> nat)                \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat)) \<Rightarrow> bool"
where
  "uses_sub v1 [] v2 sub     = (sub = (v1,v2))"
| "uses_sub v1 (a#as) v2 sub = (sub = (v1,src a) \<or> uses_sub (tgt a) as v2 sub)"

lemma uses_sub_one :
  "uses_sub v1 [arc] v2 (d,a) = (tgt arc = d \<and> v2 = a \<or> v1 = d \<and> src arc = a)"
by auto


lemma 
  assumes "cas v1 as v2 subs"
  shows   "Graph.cas v1 as v2 \<or> (\<exists> as1 as2. \<exists> (d,a) \<in> subs\<^sup>+. cas v1 as1 d subs \<and> cas a as2 v2 subs)"
proof (case_tac as)
  assume "as = []"
  hence  "v1 = v2 \<or> (v1,v2) \<in> subs\<^sup>+" using assms(1) by simp

  thus ?thesis
  proof (elim disjE)
    assume "v1 = v2" thus ?thesis using `as = []` by simp
  next
    assume "(v1,v2) \<in> subs\<^sup>+" thus ?thesis using `as = []`
    apply (intro disjI2)
    apply (rule_tac ?x="[]" in exI)
    apply (rule_tac ?x="[]" in exI)
    apply auto
oops

lemma
  assumes "uses_sub v1 (as1 @ as2) v2 (d,a)"
  shows   "cas v1 (as1 @ as2) v2 subs = (\<exists> (d,a) \<in> subs\<^sup>+. cas v1 as1 d subs \<and> cas a as2 v2 subs)"
using assms
oops




(* Soit p un sous-chemin (ac subsomp\<degree>) de l'extension de g reliant les sommets v1 et v2.
   Si p n'utilise pas la nouvelle subsomption, alors p est également un sous-chemin reliant 
   v1 et v2 dans g *)
lemma (in wf_sub_rel_of) sp_in_extends_not_using_sub :
  assumes "extends g subs (d,a) subs'"
  assumes "subpath g v1 p v2 subs'"
  assumes "\<not> uses_sub v1 p v2 (d,a)"

  shows   "subpath g v1 p v2 subs"
using assms(2) assms(3)
proof (induction p arbitrary : v1)

  case (Nil v1)
  show ?case
  proof -
    
    have "v1 \<noteq> d \<or> v2 \<noteq> a" using Nil(2) by auto

    thus ?thesis using assms(1) Nil(1) by (simp add : sp_Nil_in_extends)

  qed

next

  case (Cons arc p v1)
  show ?case
  proof -

    find_theorems name:"local" name:"Cons"

    have "(d,a) \<noteq> (v1,src arc)"
    and  "\<not> uses_sub (tgt arc) p v2 (d,a)"
    using Cons(3) by simp_all

    moreover
    have "sub_rel_of g subs'" 
    using sub_rel_of assms(1) by (rule sub_rel_of.extends_imp_sub_rel_of)

    hence "v1 = src arc \<or> (v1, src arc) \<in> subs'\<^sup>+"
    and   "arc \<in> arcs g"
    and   "subpath g (tgt arc) p v2 subs'"
    using Cons(2) sub_rel_of by (simp_all add : sub_rel_of.sp_Cons)

    ultimately
    have "subpath g (tgt arc) p v2 subs" by (intro Cons.IH)

    show ?thesis using `v1 = src arc \<or> (v1, src arc) \<in> subs'\<^sup>+`
    proof (elim disjE)

      assume "v1 = src arc"

      thus ?thesis 
      using sub_rel_of `arc \<in> arcs g` `subpath g (tgt arc) p v2 subs`
      by (simp add : sub_rel_of.sp_Cons)

    next

      assume "(v1, src arc) \<in> subs'\<^sup>+" 

      hence  "(v1, src arc) \<in> subs'" 
      using assms(1) 
            extends_imp_wf_sub_rel_of[of "(d,a)" subs'] 
            wf_sub_rel_of.in_trancl_imp[of g subs' v1 "src arc"]
      by simp
      
      moreover
      have   "(d,a) \<notin> subs" using assms(1) by (rule extends_imp_not_mem)

      ultimately
      have  "(v1, src arc) \<in> subs" 
      using `(d,a) \<noteq> (v1,src arc)` assms(1) by (auto simp add : extends_def)

      hence "(v1, src arc) \<in> subs\<^sup>+" by auto

      thus ?thesis 
      using sub_rel_of `arc \<in> arcs g` `subpath g (tgt arc) p v2 subs`
      by (simp add : sub_rel_of.sp_Cons)

   qed
 qed
qed


(* Réciproquement, si p relie v1 et v2 dans g et dans son extension, alors p n'utilise pas 
   la nouvelle subsomption. 

   On peut certainement prouver ca plus facilement sans induction.

*)

lemma (in wf_sub_rel_of) sp_in_extends_and_sp_imp_not_using_sub :
  assumes "extends g subs (d,a) subs'"
  assumes "subpath g v1 p v2 subs"

  shows   "\<not> uses_sub v1 p v2 (d,a)"  
using assms(2)
proof (induction p arbitrary : v1)

  case (Nil v1)
  show ?case
  proof (intro notI)

    assume "uses_sub v1 [] v2 (d,a)"

    hence  "(d,a) = (v1,v2)" by simp

    hence  "v1 = v2 \<or> (v1,v2) \<in> subs\<^sup>+"
    using Nil(1) by (auto simp add : subpath_def)

    thus "False"
    proof (elim disjE)
      assume "v1 = v2"

      thus ?thesis  using `(d,a) = (v1,v2)` assms(1) by (auto simp add : extends_def)
    next
      assume "(v1, v2) \<in> subs\<^sup>+"

      hence  "(v1, v2) \<in> subs" by (rule in_trancl_imp)

      thus ?thesis using `(d,a) = (v1,v2)` assms(1) by (auto simp add : extends_imp_not_mem)
    qed
  qed

next

  case (Cons arc p v1)
  show ?case
  proof (intro notI)

    assume "uses_sub v1 (arc # p) v2 (d, a)"

    hence  "(d,a) = (v1, src arc) \<or> uses_sub (tgt arc) p v2 (d,a)" by simp

    thus "False"
    proof (elim disjE)

      assume "(d,a) = (v1, src arc)"

      moreover
      have "v1 = src arc \<or> (v1, src arc) \<in> subs\<^sup>+"
      and  "arc \<in> arcs g"
      and  "subpath g (tgt arc) p v2 subs"
      using Cons(2) sub_rel_of by (simp_all add : sub_rel_of.sp_Cons)


      thus ?thesis
      proof (elim disjE)

        assume "v1 = src arc"
        thus ?thesis using assms(1) `(d,a) = (v1, src arc)` by (auto simp add : extends_def)

      next

        assume "(v1, src arc) \<in> subs\<^sup>+"

        hence "(v1, src arc) \<in> subs" by (rule in_trancl_imp)
        
        thus ?thesis using assms(1) `(d,a) = (v1, src arc)` 
        by (simp add : extends_imp_not_mem)

      qed

    next

      assume "uses_sub (tgt arc) p v2 (d, a)"

      moreover
      have "sub_rel_of g subs'" 
      using sub_rel_of assms(1) by (rule sub_rel_of.extends_imp_sub_rel_of)

      hence "subpath g (tgt arc) p v2 subs"
      using sub_rel_of Cons(2) by (simp_all add : sub_rel_of.sp_Cons)

      hence  "\<not> uses_sub (tgt arc) p v2 (d, a)" by (rule Cons.IH)

      ultimately
      show ?thesis by simp
    
    qed
  qed
qed




(*lemma (in wf_sub_rel_of) sp_in_extends_imp1 : 
  assumes "extends g subs (d,a) subs'"
  assumes "subpath g v1 p v2 subs'"
  assumes "v1 \<noteq> d"

  shows   "subpath g v1 p v2 subs \<or> (\<exists> p1 p2. p = p1 @ p2 \<and> p1 \<noteq> [] \<and> subpath g v1 p1 d subs)"
          (is "?P p v1")
proof (case_tac "uses_sub v1 p v2 (d,a)")

  assume "\<not> uses_sub v1 p v2 (d,a)"
  thus ?thesis using assms(1) assms(2) by (simp add : sp_in_extends_not_using_sub)

next

  assume "uses_sub v1 p v2 (d,a)"

  thus ?thesis
  using assms(2) assms(3)
  proof (induction p arbitrary : v1)

    case (Nil v1) 
    show ?case
    proof -
      have "v1 = d" using Nil(1) by simp
      thus ?thesis  using Nil(3) by simp
    qed

  next

    case (Cons arc p v1)
    show ?case
    proof -

      have "sub_rel_of g subs'" 
      using sub_rel_of assms(1) by (rule sub_rel_of.extends_imp_sub_rel_of)

      hence "src arc = v1 \<or> (v1, src arc) \<in> subs'"
      and   "arc \<in> arcs g" 
      and   "subpath g (tgt arc) p v2 subs'"
      using Cons(3) by (simp_all add : sub_rel_of.sp_Cons)

      hence "subpath g v1 [arc] (tgt arc) subs'" 
      using `sub_rel_of g subs'` by (simp add : sub_rel_of.sp_one)

      have "subpath g v1 [arc] (tgt arc) subs"
      proof -

        have "\<not> uses_sub v1 [arc] (tgt arc) (d,a)" using Cons(2) Cons(4) assms(1)
        by (auto simp add : uses_sub_one extends_def)

        thus ?thesis
        using assms(1) `subpath g v1 [arc] (tgt arc) subs'`
        by (elim sp_in_extends_not_using_sub)

      qed

      thus ?thesis
      proof (case_tac "tgt arc = d")
        
        assume "tgt arc = d"

        thus ?thesis 
        using `subpath g v1 [arc] (tgt arc) subs` 
        by (intro disjI2, rule_tac ?x="[arc]" in exI) auto

      next

        assume "tgt arc \<noteq> d"

        moreover
        have "uses_sub (tgt arc) p v2 (d,a)" using Cons(2) Cons(4) by simp

        ultimately
        have "?P p (tgt arc)" using `subpath g (tgt arc) p v2 subs'` by (intro Cons.IH)

        thus ?thesis
        proof (elim disjE)

          assume "subpath g (tgt arc) p v2 subs"

          hence  "\<not> uses_sub (tgt arc) p v2 (d,a)"
          using wf_sub_rel_of assms(1) `subpath g (tgt arc) p v2 subs'` 
          by (intro sp_in_extends_and_sp_imp_not_using_sub)

          thus ?thesis using `uses_sub (tgt arc) p v2 (d,a)` by simp

        next

          assume "\<exists>p1 p2. p = p1 @ p2 \<and> p1 \<noteq> [] \<and> subpath g (tgt arc) p1 d subs"

          then obtain p1 p2 where "p = p1 @ p2"
                            and   "p1 \<noteq> []"
                            and   "subpath g (tgt arc) p1 d subs"
          by blast

          hence "subpath g v1 (arc # p1) d subs"
          using `subpath g v1 [arc] (tgt arc) subs` sub_rel_of by (simp add : sub_rel_of.sp_Cons)

          thus ?thesis using `p = p1 @ p2`
          by (intro disjI2, rule_tac ?x="arc # p1" in exI) auto
        qed
      qed
    qed
  qed
qed*)


(* Dans le cas où v1 = d, soit p = [] soit p est un sous-chemin depuis a *)
lemma (in wf_sub_rel_of) sp_in_extends_imp2 :
  assumes "extends g subs (d,a) subs'"
  assumes "subpath g d p v subs'"
  
  shows   "p = [] \<or> subpath g a p v subs'"
proof -

  have "(d,a) \<in> subs'"  
  and  "out_arcs g d = {}" using assms(1) by (simp_all add : extends_def)

  moreover
  have "wf_sub_rel_of g subs'" using assms(1) by (rule extends_imp_wf_sub_rel_of)

  ultimately
  show ?thesis using assms by (elim wf_sub_rel_of.sp_from_subsumee)
  thm wf_sub_rel_of.sp_from_subsumee
qed




lemma (in wf_sub_rel_of) sp_in_extends_imp3 :
  assumes "extends g subs (d,a) subs'"
  assumes "subpath_from g d as subs'"
  
  shows   "subpath_from g a as subs'"
proof -
  have "(d,a) \<in> subs'"  
  and  "out_arcs g d = {}" using assms(1) by (simp_all add : extends_def)

  moreover
  have "wf_sub_rel_of g subs'" using assms(1) by (rule extends_imp_wf_sub_rel_of)

  ultimately
  show ?thesis using assms by (elim wf_sub_rel_of.sp_from_subsumee2)
qed




(* Un sous-chemin p de l'extension de g par (d,a) reliant v1 à v2 peut être décomposé en deux 
   sous-chemins p1 et p2 tels que :
   - p1 est non-vide et est un sous-chemin de g menant de v1 à d,
   - p2 est un sous-chemin de l'extension de g menant de d à v2.

   Note : on a besoin d'éliminer le cas où v1 = d car dans ce cas on ne peut pas trouver un p1
   n'utilisant pas (d,a)
*)
lemma (in wf_sub_rel_of) sp_in_extends_imp : 
  assumes "extends g subs (d,a) subs'"
  assumes "subpath g v1 p v2 subs'"
  assumes "v1 \<noteq> d"

  shows   "subpath g v1 p v2 subs \<or> 
           (\<exists> p1 p2. p = p1 @ p2 \<and> p1 \<noteq> [] \<and> subpath g v1 p1 d subs \<and> subpath g d p2 v2 subs')"
          (is "?P p v1")
proof (case_tac "uses_sub v1 p v2 (d,a)")

  assume "\<not> uses_sub v1 p v2 (d,a)"
  thus ?thesis using assms(1,2) by (simp add : sp_in_extends_not_using_sub)

next

  assume "uses_sub v1 p v2 (d,a)"

  thus ?thesis
  using assms(2) assms(3)
  proof (induction p arbitrary : v1)

    case (Nil v1) 
    show ?case
    proof -
      have "v1 = d" using Nil(1) by simp
      thus ?thesis  using Nil(3) by simp
    qed

  next

    case (Cons arc p v1)
    show ?case
    proof -

      have "sub_rel_of g subs'" 
      using sub_rel_of assms(1) by (rule sub_rel_of.extends_imp_sub_rel_of)

      hence "v1 = src arc \<or> (v1, src arc) \<in> subs'\<^sup>+"
      and   "arc \<in> arcs g" 
      and   "subpath g (tgt arc) p v2 subs'"
      using Cons(3) by (simp_all add : sub_rel_of.sp_Cons)

      hence "subpath g v1 [arc] (tgt arc) subs'" 
      using `sub_rel_of g subs'` by (simp add : sub_rel_of.sp_one)

      have "subpath g v1 [arc] (tgt arc) subs"
      proof -

        have "\<not> uses_sub v1 [arc] (tgt arc) (d,a)" using Cons(2) Cons(4) assms(1)
        by (auto simp add : uses_sub_one extends_def)

        thus ?thesis
        using assms(1) `subpath g v1 [arc] (tgt arc) subs'`
        by (elim sp_in_extends_not_using_sub)

      qed

      thus ?thesis
      proof (case_tac "tgt arc = d")
        
        assume "tgt arc = d"

        thus ?thesis 
        using `subpath g v1 [arc] (tgt arc) subs`  `subpath g (tgt arc) p v2 subs'`
        by (intro disjI2, rule_tac ?x="[arc]" in exI) auto

      next

        assume "tgt arc \<noteq> d"

        moreover
        have "uses_sub (tgt arc) p v2 (d,a)" using Cons(2) Cons(4) by simp

        ultimately
        have "?P p (tgt arc)" using `subpath g (tgt arc) p v2 subs'` by (intro Cons.IH)

        thus ?thesis
        proof (elim disjE)

          assume "subpath g (tgt arc) p v2 subs"

          hence  "\<not> uses_sub (tgt arc) p v2 (d,a)"
          using wf_sub_rel_of assms(1) `subpath g (tgt arc) p v2 subs'` 
          by (intro sp_in_extends_and_sp_imp_not_using_sub)

          thus ?thesis using `uses_sub (tgt arc) p v2 (d,a)` by simp

        next

          assume "\<exists>p1 p2. p = p1 @ p2 \<and> p1 \<noteq> [] 
                        \<and> subpath g (tgt arc) p1 d subs 
                        \<and> subpath g d p2 v2 subs'"

          then obtain p1 p2 where "p = p1 @ p2"
                            and   "p1 \<noteq> []"
                            and   "subpath g (tgt arc) p1 d subs"
                            and   "subpath g d p2 v2 subs'"
          by blast

          hence "subpath g v1 (arc # p1) d subs"
          using `subpath g v1 [arc] (tgt arc) subs` sub_rel_of by (simp add : sub_rel_of.sp_Cons)

          thus ?thesis using `p = p1 @ p2` `subpath g d p2 v2 subs'`
          by (intro disjI2, rule_tac ?x="arc # p1" in exI) auto
        qed
      qed
    qed
  qed
qed


(*** Note : avec les deux lemmes précédents, on a que tout sous-chemin de l'extension qui n'est pas 
     issu de d peut être décomposé en un nombre fini de sous-chemins du "graphe" (du couple graphe +
     relation de subsomption avant extension) éventuellement succédés de [] allant de d à d (c'est 
     un défaut de la modélisation des séquences d'arcs consistantes, etc).

     J'espère par la suite me servir de cette propriété pour prouver les lemmes décrivant les 
     ensembes de chemins dans la structure avec exécution symbolique.

 *)


lemma (in wf_sub_rel_of) extends_sp_from_set : 
  assumes "extends g subs (d,a) subs'"
  assumes "v1 \<noteq> d"

  shows   
    "subpaths_from g v1 subs' = subpaths_from g v1 subs 
                              \<union> {p1 @ p2 | p1 p2 v2. p1 \<noteq> [] 
                                                   \<and> subpath g v1 p1 d subs 
                                                   \<and> subpath g d p2 v2 subs'}"
  (is "?L = ?R1 \<union> ?R2")
proof (subst set_eq_iff, intro allI iffI)

  fix p

  assume "p \<in> ?L"

  then obtain v2 where "subpath g v1 p v2 subs'" unfolding subpath_from_def by blast

  hence  "subpath g v1 p v2 subs \<or> (\<exists> p1 p2. p = p1 @ p2 
                                           \<and> p1 \<noteq> [] 
                                           \<and> subpath g v1 p1 d subs 
                                           \<and> subpath g d p2 v2 subs')"
  using assms by (elim sp_in_extends_imp)

  thus "p \<in> ?R1 \<union> ?R2"
  apply (simp only  : subpath_from_def Un_iff mem_Collect_eq)
  apply (elim disjE)
    apply (intro disjI1, blast)
    apply (intro disjI2, blast)
  done

next

  fix p

  assume "p \<in> ?R1 \<union> ?R2"
 
  thus   "p \<in> ?L"
  proof (subst (asm) Un_iff, elim disjE)

    assume "p \<in> ?R1"

    thus ?thesis using assms(1) unfolding subpath_from_def mem_Collect_eq
    apply (elim exE)
    apply (rule_tac ?x="v'" in exI)
    apply (rule sp_in_extends)
      apply assumption
      apply assumption
    done

  next

    assume "p \<in> ?R2"

    then obtain p1 p2 v2 where "p = p1 @ p2"
                         and   "p1 \<noteq> []"
                         and   "subpath g v1 p1 d subs"
                         and   "subpath g d p2 v2 subs'"
    by blast

    moreover
    hence "subpath g v1 p1 d subs'"
    using assms(1) by (elim sp_in_extends)
    
    moreover
    have "wf_sub_rel_of g subs'" using assms(1) by (rule extends_imp_wf_sub_rel_of)

    ultimately
    have "subpath g v1 (p1 @ p2) v2 subs'"
    by (simp only : wf_sub_rel_of.sp_append) blast
    
    thus ?thesis using `p = p1 @ p2` by (simp only : subpath_from_def) blast

  qed

qed







lemma (in wf_sub_rel_of) extends_sp_from_set2 : 
  assumes "extends g subs (d,a) subs'"

  shows   
    "subpaths_from g d subs' = {[]} \<union> subpaths_from g a subs'"
  (is "?L = ?R")

proof (simp only : set_eq_iff, intro allI iffI)

  fix p
  
  assume "p \<in> ?L"

  then obtain v where "subpath g d p v subs'" unfolding subpath_from_def by blast

  hence "p = [] \<or> subpath g a p v subs'" using assms(1) by (elim sp_in_extends_imp2)
  thm sp_in_extends_imp2

  thus "p \<in> ?R" by (auto simp only : subpath_from_def)

next

  fix p

  assume "p \<in> ?R"

  hence  "p = [] \<or> p \<in> subpaths_from g a subs'" by simp

  thus   "p \<in> ?L"
  proof (elim disjE)

    assume "p = []"

    moreover
    have "d \<in> Graph.vertices g" using assms(1) by (auto simp add : extends_def)

    ultimately
    show ?thesis
    apply (simp only : subpath_from_def mem_Collect_eq)
    apply (rule_tac ?x="d" in exI)
    apply (simp add : Nil_sp)
    done

  next

    assume "p \<in> subpaths_from g a subs'"

    then obtain v where  "subpath g a p v subs'" unfolding subpath_from_def by blast

    moreover
    have "wf_sub_rel_of g subs'" using assms(1) by (rule extends_imp_wf_sub_rel_of)

    moreover
    have "(d,a) \<in> subs'" using assms(1) by (simp add : extends_def)

    ultimately
    have "subpath g d p v subs'" by (elim wf_sub_rel_of.asp_imp_dsp)
    
    thus ?thesis unfolding subpath_from_def by blast
    
  qed
qed


lemma (in wf_sub_rel_of) extends_sp_from_set3 : 
  assumes "extends g subs (d,a) subs'"

  shows   
    "subpaths_from g d subs' = subpaths_from g a subs'"
  (is "?L = ?R")

proof (simp only : set_eq_iff, intro allI iffI)

  fix as 

  assume "as \<in> ?L"

  thus   "as \<in> ?R" using assms by (simp add : sp_in_extends_imp3)

next

  fix as

  assume "as \<in> ?R"

  moreover
  have "wf_sub_rel_of g subs'" using assms by (elim extends_imp_wf_sub_rel_of)

  moreover
  have "(d,a) \<in> subs'" using assms extendsE2 by fast

  ultimately
  show   "as \<in> ?L"
  using  wf_sub_rel_of.asp_imp_dsp unfolding subpath_from_def by blast
  
qed



find_unused_assms

lemma (in wf_sub_rel_of) sp_from_subsumer_imp :
  assumes "sub \<in> subs"
  assumes "subpath g (subsumer sub) as v subs"
  shows   "subpath g (subsumee sub) as v subs"
proof (cases as)
  
  case goal1

  hence "subsumer sub = v \<or> (subsumer sub,v) \<in> subs\<^sup>+" using assms(2) by (simp add : Nil_sp)

  thus ?thesis
  proof (elim disjE)
    assume "subsumer sub = v"

    moreover
    have   "subsumee sub \<in> Graph.vertices g"
    using assms(1) sub_rel_of unfolding sub_rel_of_def subsumees_def by blast

    moreover
    have   "(subsumee sub, subsumer sub) \<in> subs" using assms(1) by simp

    ultimately
    show ?thesis using goal1 by (simp add : Nil_sp r_into_trancl)
  next
    assume "(subsumer sub, v) \<in> subs\<^sup>+"

    hence  "(subsumer sub, v) \<in> subs" by (rule in_trancl_imp)

    moreover
    have   "(subsumee sub, subsumer sub) \<in> subs" using assms(1) by simp

    ultimately
    have  "subsumers subs \<inter> subsumees subs \<noteq> {}" 
    unfolding subsumers_conv subsumees_conv by blast
    
    thus ?thesis 
    using wf_sub_rel unfolding wf_sub_rel_def by simp
  qed

next

  case (goal2 a as')

  hence "subsumer sub = src a \<or> (subsumer sub,src a) \<in> subs\<^sup>+" 
  using assms(2) sub_rel_of by (simp add : sub_rel_of.sp_Cons)

  thus ?thesis
  proof (elim disjE)
  
    assume "subsumer sub = src a"

    hence "(subsumee sub, src a) \<in> subs\<^sup>+"
    using assms(1) sub_surj_pairing[of sub] by (simp add : r_into_trancl')

    thus ?thesis using assms goal2 sub_rel_of by (simp add : sub_rel_of.sp_Cons)

  next

    assume "(subsumer sub, src a) \<in> subs\<^sup>+"
    
    hence  "(subsumer sub, src a) \<in> subs" by (rule in_trancl_imp)

    moreover
    have   "(subsumee sub, subsumer sub) \<in> subs" using assms(1) by simp

    ultimately
    have  "subsumers subs \<inter> subsumees subs \<noteq> {}" 
    unfolding subsumers_conv subsumees_conv by blast
    
    thus ?thesis using wf_sub_rel unfolding wf_sub_rel_def by simp
  qed
qed


end
