theory Tree
imports ArcExt
begin

(**************************************************************************************************)
(***************************  CARACTERISATION DES ARBRES  *****************************************)
(**************************************************************************************************)

(* Un graphe est un arbre si pour chacun de ses sommets, il existe un unique chemin y menant *)

locale is_tree = rgraph +
  assumes tree_is : "l \<in> Graph.vertices g \<Longrightarrow> \<exists>! p. Graph.path g p l"


(* le graphe vide est un arbre *)
lemma empty_graph_is_tree :
  assumes "arcs g = {}"
  shows   "is_tree g"
using assms arcs_empty_imp_path_eq[of g] 
      empty_graph_vertices[of g] 
unfolding is_tree_def by auto


(* si deux noeuds d'un arbre sont reliés par un sous-chemin, ce sous-chemin est unique *)
lemma (in is_tree) ex1_path_to_vert :
  assumes "Graph.subpath g l1 p l2"
  shows   "\<forall> p'. Graph.subpath g l1 p' l2 \<longrightarrow> p' = p"
proof -           
  from assms(1)
  have l1_vert : "l1 \<in> Graph.vertices g"
  and  l2_vert : "l2 \<in> Graph.vertices g" 
  by (auto simp add : Graph.fst_of_sp_is_vert Graph.lst_of_sp_is_vert)

  then obtain p_l1 where "Graph.path g p_l1 l1"
                   and   "\<forall> p'. Graph.path g p' l1 \<longrightarrow> p' = p_l1"
  using tree_is by blast

  hence  "Graph.path g (p_l1 @ p) l2" using `Graph.subpath g l1 p l2` by (intro Graph.path_appendI)

  hence "\<forall> p'. Graph.path g p' l2 \<longrightarrow> p' = (p_l1 @ p)" using tree_is l2_vert by auto

  show ?thesis
  proof (rule ccontr, simp only : not_all not_imp)
    
    assume "\<exists>p'. Graph.subpath g l1 p' l2 \<and> p' \<noteq> p"

    then obtain p' where "Graph.subpath g l1 p' l2"
                   and   "p' \<noteq> p"
    by blast

    hence "Graph.path g (p_l1 @ p') l2" 
    and   "p_l1 @ p' \<noteq> p_l1 @ p" 
    using `Graph.path g p_l1 l1` by (auto simp add : Graph.path_appendI)

    thus "False" using `\<forall> p'. Graph.path g p' l2 \<longrightarrow> p' = (p_l1 @ p)` by simp
  qed
qed

(* soient g un arbre et l un noeud de cet arbre, il existe un unique chemin menant à l dans 
   l'extension (par ajout d'arc) de g par a *)
lemma (in is_tree) path_imp_unique_in_extends :
  assumes "extends g a g'"
  assumes "l \<in> Graph.vertices g"
  shows   "\<exists>! p. Graph.path g' p l"
proof -
  from tree_is assms(2)
  obtain p where "Graph.path g p l"
             and "\<forall> p'. Graph.path g p' l \<longrightarrow> p' = p"
  unfolding Ex1_def by blast

  have  "Graph.path g' p l"
  using `Graph.path g p l` using assms(1) by (elim path_in_extends)
  
  moreover
  have "\<forall> p'. Graph.path g' p' l \<longrightarrow> p' = p"
  proof (intro allI impI)
    fix p'

    assume p'_path : "Graph.path g' p' l"

    show "p' = p"
    proof (case_tac "a \<in> set p'")

      assume "a \<in> set p'"

      moreover
      have "out_arcs g' (tgt a) = {}" 
      using assms(1)  by (rule extends_tgt_out_arcs)

      ultimately
      have "\<exists>p''. p' = p'' @ [a] \<and> a \<notin> set p''" 
      using p'_path by (intro path_through_de_decomp)

      then obtain p'' where "p' = p'' @ [a] \<and> a \<notin> set p''" by blast

      hence "l = tgt a" using p'_path by (simp add : path_append_one)

      thus ?thesis 
      proof -
        have "l \<in> Graph.vertices g" using `Graph.path g p l` by (rule lst_of_path_is_vert)

        thus ?thesis using `l = tgt a` assms(1) extendsE2[of g a g'] by simp
      qed
    
    next
      assume "a \<notin> set p'"

      hence  "Graph.path g p' l" 
      using assms p'_path by (simp add : p_in_extends_minus_arc_imp_p_in_old)

      thus ?thesis using `\<forall> p'. Graph.path g p' l \<longrightarrow> p' = p` by simp
    qed
  qed

  ultimately
  show ?thesis by auto
qed


(* il existe un unique chemin menant à la cible de l'arc a dans l'extension d'un graphe g par a *)
lemma (in is_tree) unique_p_to_tgt_in_extends :
  assumes "extends g a g'"
  shows   "\<exists>! p. Graph.path g' p (tgt a)"
proof -
  from tree_is assms(1) extendsE1[of g a g']
  obtain p where "Graph.path g p (src a)"
           and   "\<forall> p'. Graph.path g p' (src a) \<longrightarrow> p' = p"
  by blast

  have  "Graph.path g' p (src a)"  using assms(1) `Graph.path g p (src a)` by (rule path_in_extends)

  hence "Graph.path g' (p @ [a]) (tgt a)" 
  using assms(1) extendsE3[of g a g'] by (auto simp add : path_append_one add_arc_arcs)
  
  moreover
  have "\<forall> p'. Graph.path g' p' (tgt a) \<longrightarrow> p' = p @ [a]"
  proof (intro allI impI)
    fix p'

    assume "Graph.path g' p' (tgt a)"

    moreover
    hence   "a \<in> set p'" using assms by (elim p_extends_ends_in_tgt_imp_mem)

    moreover
    have   "out_arcs g' (tgt a) = {}" using assms by (rule extends_tgt_out_arcs)

    ultimately
    have "\<exists>p''. p' = p'' @ [a] \<and> a \<notin> set p''" by (elim path_through_de_decomp)

    then obtain p'' where "p' = p'' @ [a]" and "a \<notin> set p''" by blast

    hence "Graph.path g' p'' (src a)" using `Graph.path g' p' (tgt a)` 
    by (auto simp add : path_append_one)

    hence "Graph.path g p'' (src a)"
    using assms(1) extendsE1[of g a g'] `a \<notin> set p''` 
    by (simp add : p_in_extends_minus_arc_imp_p_in_old)

    hence "p'' = p" using `\<forall> p'. Graph.path g p' (src a) \<longrightarrow> p' = p` by simp

    thus "p' = p @ [a]" using `p' = p'' @ [a]` by simp
  qed
  
  ultimately
  show ?thesis by auto
qed


(* finalement, on peut montrer que l'extension (par ajout d'arc) d'un arbre est également un arbre 
   (pas montré pour le cas par ajout de subsomption mais trivial puisque ce n'est pas le graphe qui
   est modifié mais la relation) *)
lemma (in is_tree) extends_is_tree :
  assumes "extends g a g'"
  shows   "is_tree g'"
unfolding is_tree_def
proof (intro allI impI)
  fix l

  assume "l \<in> Graph.vertices g'"

  hence  "l \<in> Graph.vertices g \<or> l = tgt a" using assms(1) by (auto simp add : extends_vertices)

  thus   "\<exists>!p. Graph.path g' p l"
  proof (elim disjE)
    assume "l \<in> Graph.vertices g"
    
    thus ?thesis using assms by (elim path_imp_unique_in_extends)
    thm path_imp_unique_in_extends
  next
    assume "l = tgt a"

    thus ?thesis using assms by (simp add : unique_p_to_tgt_in_extends)
  qed
qed

lemma (in is_tree) sp_between_self :
  assumes "Graph.subpath g l as l"
  shows   "as = []"
proof -
  from assms have "l \<in> Graph.vertices g" by (rule Graph.fst_of_sp_is_vert)

  hence "Graph.subpath g l [] l" by (simp add : Graph.Nil_sp)

  thus ?thesis using assms ex1_path_to_vert by fast
qed



lemma (in is_tree)
  assumes "finite (arcs g)"
  shows   "finite (Graph.paths g)"
proof -

oops

find_unused_assms

end
